/**
 *
 * Unity/Compiz Dialog Handler Plugin
 *
 * unitydialog.cpp
 *
 * Copyright : (C) 2010 Canonical Ltd.
 *
 * Authors: Sam Spilsbury   <sam.spilsbury@canonical.com>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 **/

#include "unitydialog.h"

COMPIZ_PLUGIN_20090315 (unitydialog, UnityDialogPluginVTable);

void
UnityDialogScreen::cairoClear (cairo_t    *cr)
{
    if (pixmap)
    {
        XFreePixmap (screen->dpy (), pixmap);
	pixmap = None;
    }
    if (surface)
    {
	cairo_surface_destroy (surface);
	surface = NULL;
    }
    if (cairo)
    {
	cairo_destroy (cairo);
	cairo = NULL;
    }

    cairo = cairoContext ();

    if (cairo)
    {
	cairo_save (cairo);
	cairo_set_operator (cairo, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cairo);
	cairo_restore (cairo);
    }
}

cairo_t *
UnityDialogScreen::cairoContext ()
{
    if (!cairo)
    {
	XRenderPictFormat *format;
	Screen		  *xScreen;
	int		  w, h;

	xScreen = ScreenOfDisplay (screen->dpy (), screen->screenNum ());

	w = 4;
	h = 4;

	format = XRenderFindStandardFormat (screen->dpy (),
					    PictStandardARGB32);

	pixmap = XCreatePixmap (screen->dpy (), screen->root (), w, h, 32);

	texture = GLTexture::bindPixmapToTexture (pixmap, w, h, 32);

	if (texture.empty ())
	{
	    compLogMessage ("annotate", CompLogLevelError,
			    "Couldn't bind pixmap 0x%x to texture",
			    (int) pixmap);

	    XFreePixmap (screen->dpy (), pixmap);

	    return NULL;
	}

	damage = XDamageCreate (screen->dpy (), pixmap,
				XDamageReportRawRectangles);

	surface =
	    cairo_xlib_surface_create_with_xrender_format (screen->dpy (),
							   pixmap, xScreen,
							   format, w, h);

	cairo = cairo_create (surface);
    }

    return cairo;
}

void
UnityDialogScreen::renderDim ()
{
    cairoClear (cairo);

    cairo_set_line_width (cairo, 2);
    cairo_set_source_rgba (cairo,
			   (optionGetColorRed () / 65535.0f),
			   (optionGetColorBlue () / 65535.0f),
			   (optionGetColorGreen () / 65535.0f),
			   (optionGetColorAlpha () / (65535.0f)));

    cairo_move_to (cairo, 0, 0);
    cairo_rectangle (cairo, 0, 0, 4, 4);

    cairo_fill (cairo);
}

/* Draw the window */

bool
UnityDialogWindow::glDraw (const GLMatrix &transform,
		   GLFragment::Attrib &fragment,
		   const CompRegion &region,
		   unsigned int mask)
{
    /* We want to set the geometry of the dim to the window
     * region */
    CompRegion reg = CompRegion (window->x (), window->y (), window->width (), window->height ());

    /* Draw the window on the bottom, we will be drawing the
     * dim render on top */
    bool status = gWindow->glDraw (transform, fragment, region, mask);

    UNITY_DIALOG_SCREEN (screen);

    foreach (GLTexture *tex, uds->texture)
    {
	GLTexture::MatrixList matl;
	GLTexture::Matrix     mat = tex->matrix ();

	/* We can reset the window geometry since it will be
	 * re-added later */
	gWindow->geometry ().reset ();

	/* Scale the dim render by the ratio of vignette size
	 * to window size */
	mat.xx *= (4) / window->width ();
	mat.yy *= (4) / window->height ();

	/* Not sure what this does, but it is necessary
	 * (adjusts for scale?) */
	mat.x0 -= mat.xx * reg.boundingRect ().x1 ();
	mat.y0 -= mat.yy * reg.boundingRect ().y1 ();

	matl.push_back (mat);

	/* Now allow plugins to mess with the geometry of our
	 * vignette (so we get a nice render for things like
	 * wobbly etc etc */
	gWindow->glAddGeometry (matl, reg, infiniteRegion);

	/* Did it succeed? */
	if (gWindow->geometry ().vertices)
	{
	    /* Texture rendering set-up */
	    uds->gScreen->setTexEnvMode (GL_MODULATE);
	    glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	    /* Draw the vignette texture with all of it's modified
	     * geometry glory */
	    gWindow->glDrawTexture (tex, fragment, mask | PAINT_WINDOW_BLEND_MASK
				    | PAINT_WINDOW_TRANSLUCENT_MASK |
				      PAINT_WINDOW_TRANSFORMED_MASK);
	    /* Texture rendering tear-down */
	    glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	    uds->gScreen->setTexEnvMode (GL_REPLACE);
	}
     }

    return status;
}

void
UnityDialogWindow::addTransient (CompWindow *w)
{
    bool alreadyAdded = false;

    if (!mTransients.size ())
    {
	gWindow->glDrawSetEnabled (this, true);
	window->grabNotifySetEnabled (this, true);
	window->ungrabNotifySetEnabled (this, true);
	window->moveNotifySetEnabled (this, true);
	cWindow->addDamage ();
    }

    foreach (CompWindow *tw, mTransients)
	if (tw->id () == w->id ())
	    alreadyAdded = true;

    if (!alreadyAdded)
    {
	w->grabNotifySetEnabled (this, true);
	w->ungrabNotifySetEnabled (this, true);
	w->moveNotifySetEnabled (this, true);

	UnityDialogWindow::get (w)->mParent = window;

	mTransients.push_back (w);
    }
}

void
UnityDialogWindow::removeTransient (CompWindow *w)
{
    mTransients.remove (w);

    w->grabNotifySetEnabled (this, false);
    w->ungrabNotifySetEnabled (this, false);
    w->moveNotifySetEnabled (this, false);

    if (!mTransients.size ())
    {
	gWindow->glDrawSetEnabled (this, false);
	window->ungrabNotifySetEnabled (this, false);
	window->grabNotifySetEnabled (this, false);
	window->moveNotifySetEnabled (this, false);

	UnityDialogWindow::get (w)->mParent = NULL;

	cWindow->addDamage ();
    }
}

void
UnityDialogWindow::windowNotify (CompWindowNotify n)
{
    UnityDialogScreen *uds = UnityDialogScreen::get (screen);

    switch (n)
    {
	case CompWindowNotifyMap:

	    /* If this window is a transient for some other window,
	     * increment the transient count and
	     * enable the dimming on that other window */

	    if (window->transientFor ())
	    {
		CompWindow *parent = screen->findWindow (window->transientFor ());

		if (parent &&
		    uds->optionGetMatch ().evaluate (window))
		    UnityDialogWindow::get (parent)->addTransient (window);
	    }

	    break;

	case CompWindowNotifyUnmap:

	    /* If this window was a transient for some other window
	     * then decrement the transient count */

	    if (window->transientFor ())
	    {
		CompWindow *parent = screen->findWindow (window->transientFor ());

		if (parent &&
		    uds->optionGetMatch ().evaluate (window))
		    UnityDialogWindow::get (parent)->removeTransient (window);
	    }

	    break;

	default:
	    break;
    }

    window->windowNotify (n);
}

void
UnityDialogWindow::grabNotify (int x, int y,
			       unsigned int state,
			       unsigned int mask)
{
    window->grabNotify (x, y, state, mask);

    if (!mSkipNotify && (mask & CompWindowGrabMoveMask))
    {
	if (mParent)
	{
	    UnityDialogWindow::get (mParent)->mSkipNotify = true;
	    mParent->grabNotify (x, y, state, mask);
	    UnityDialogWindow::get (mParent)->mSkipNotify = false;
	}
	else if (mTransients.size ())
	{
	    foreach (CompWindow *w, mTransients)
	    {
		UnityDialogWindow::get (w)->mSkipNotify = true;
		w->grabNotify (x, y, state, mask);
		UnityDialogWindow::get (w)->mSkipNotify = false;
	    }
	}
    }
}

void
UnityDialogWindow::ungrabNotify ()
{
    if (!mSkipNotify)
    {
	if (mParent)
	{
	    UnityDialogWindow::get (mParent)->mSkipNotify = true;
	    mParent->ungrabNotify ();
	    mParent->syncPosition ();
	    UnityDialogWindow::get (mParent)->mSkipNotify = false;
	}
	else if (mTransients.size ())
	{
	    foreach (CompWindow *w, mTransients)
	    {
		UnityDialogWindow::get (w)->mSkipNotify = true;
		w->ungrabNotify ();
		w->syncPosition ();
		UnityDialogWindow::get (w)->mSkipNotify = false;
	    }
	}
    }

    window->ungrabNotify ();
}

void
UnityDialogWindow::moveNotify (int x, int y, bool immediate)
{
    window->moveNotify (x, y, immediate);

    if (!mSkipNotify && window->grabbed ())
    {
	if (mParent)
	{
	    UnityDialogWindow::get (mParent)->mSkipNotify = true;
	    mParent->move (x, y, immediate);
	    UnityDialogWindow::get (mParent)->mSkipNotify = false;
	}
	else if (mTransients.size ())
	{
	    foreach (CompWindow *w, mTransients)
	    {
		UnityDialogWindow::get (w)->mSkipNotify = true;
		w->move (x, y, immediate);
		UnityDialogWindow::get (w)->mSkipNotify = false;
	    }
	}
    }
}

void
UnityDialogScreen::handleEvent (XEvent *event)
{
    CompWindow *w;

    screen->handleEvent (event);

    switch (event->type)
    {
	case ClientMessage:

	    w = screen->findWindow (event->xclient.window);

	    if (event->xclient.message_type == compizResizeAtom && w)
	    {
		UnityDialogWindow *udw = UnityDialogWindow::get (w);
		CompRect parentRect (event->xclient.data.l[0],
				     event->xclient.data.l[1],
				     event->xclient.data.l[2],
				     event->xclient.data.l[3]);

		foreach (CompWindow *cw, udw->mTransients)
		{
		    int centeredX = parentRect.x () + (parentRect.width () / 2 -
						       cw->width () / 2);
		    int centeredY = parentRect.y () + (parentRect.height () / 2 -
						       cw->height () / 2);

		    cw->move (centeredX - cw->x (),
			      centeredY - cw->y (),
			      true);
		}
	    }
	default:
	    break;
    }
}

void
UnityDialogScreen::optionChanged (CompOption *opt,
				  Options    num)
{
    switch (num)
    {
	case UnitydialogOptions::Color:
	    renderDim ();
	    break;
	default:
	    break;
    }
}

/* Constructors */

UnityDialogWindow::UnityDialogWindow (CompWindow *w) :
    PluginClassHandler <UnityDialogWindow, CompWindow> (w),
    window (w),
    cWindow (CompositeWindow::get (w)),
    gWindow (GLWindow::get (w)),
    mSkipNotify (false),
    mParent (NULL)
{
    WindowInterface::setHandler (window);
    GLWindowInterface::setHandler (gWindow, false);

    window->windowNotifySetEnabled (this, true);
}

UnityDialogWindow::~UnityDialogWindow ()
{
}

UnityDialogScreen::UnityDialogScreen (CompScreen *s) :
    PluginClassHandler <UnityDialogScreen, CompScreen> (s),
    cScreen (CompositeScreen::get (s)),
    gScreen (GLScreen::get (s)),
    pixmap (None),
    surface (NULL),
    cairo (NULL),
    compizResizeAtom (XInternAtom (screen->dpy (),
				   "_COMPIZ_RESIZE_NOTIFY", 0))
{
    ScreenInterface::setHandler (screen);

    renderDim ();

    optionSetColorNotify (boost::bind (&UnityDialogScreen::optionChanged, this,
					 _1, _2));
};

UnityDialogScreen::~UnityDialogScreen ()
{
    cairoClear (cairo);
}

bool
UnityDialogPluginVTable::init ()
{
    if (!CompPlugin::checkPluginABI ("core", CORE_ABIVERSION) ||
	!CompPlugin::checkPluginABI ("composite", COMPIZ_COMPOSITE_ABI) ||
	!CompPlugin::checkPluginABI ("opengl", COMPIZ_OPENGL_ABI))
	return false;

    return true;
}
